package org.ninjasoft.cryptoslam.data;

import java.util.*;

public class Cryptogram {
    private Random random = new Random();
    private String cyphertext;
    private String plaintext;
    private char alphabet[] = new char[26];
    private char textcolor[] = new char[26];
    private static final String SAMPLE_TEXT = 
    "Welcome to Cryptoslam!  This program has been designed to help you solve\n" +
    "those pencil-and-paper cryptograms found in newspapers and puzzle \n" +
    "magazines.  First, click your mouse within this applet so that it gets \n" +
    "focus.  Next, hit a key from the menu bar above.  S,U,R let you change the\n" +
    "substitution values.  M lets you create, load, or save a message.  S gives you\n" +
    "good statistics.";
    
    public Cryptogram() {
        cyphertext = "";
        plaintext = "";
        reset();        
    }

    public void reset() {
        cyphertext = SAMPLE_TEXT;
        plaintext = "";
        for (int i=0; i<alphabet.length; i++)
            alphabet[i] = ' ';
        calcPlaintext();
    }

    public void set(char cypherLetter, char plainLetter) {
        // Normalize to uppercase
        cypherLetter = normalize(cypherLetter);
        plainLetter = normalize(plainLetter);
        if (isKey(cypherLetter))
            alphabet[cypherLetter - 'A'] = plainLetter;
    }

    public void unset(char cypherLetter) {
        set(cypherLetter, ' ');
    }
    
    public void unsetAll() {
        for (int i=0; i<alphabet.length; i++)
            alphabet[i] = ' ';
    }
    
    public void setMessage(String message) {
        this.cyphertext = message;
        calcPlaintext();
    }
    
    public String getCyphertext() {
        return this.cyphertext;
    }

    public void randomizeKeys() {
        // Probably a better way of doing this....
        int r1, r2;
        char temp;
        for (int i=0; i<26; i++)
            alphabet[i] = (char) (i + 'A');
        for (int i=0; i<1000; i++)
        {
            r1 = Math.abs(random.nextInt()) % 26;
            r2 = Math.abs(random.nextInt()) % 26;
            temp = alphabet[r1];
            alphabet[r1] = alphabet[r2];
            alphabet[r2] = temp;
        }
    }

    private void randomizeMessage() {
        boolean matchKey;
        do{
            matchKey = false;
            randomizeKeys();
            for (int i=0; i<26; i++)
            {
                if (alphabet[i] == i+'A')
                    matchKey = true;
            }
        }while(matchKey);
        calcPlaintext();
        cyphertext = plaintext;
        plaintext = "";
        for (int i=0; i<cyphertext.length(); i++)
            plaintext += " ";
        calcPlaintext();
        for (int i=0; i<26; i++)
            alphabet[i] = ' ';
    }

    private void calcPlaintext() {
        char ch;
        char p[];
        int max = cyphertext.length();
        if (max != plaintext.length()) {
            plaintext = "";
            for (int i=0; i<max; i++)
                plaintext += " ";
        }
        p = cyphertext.toCharArray();
        for (int i=0; i<max; i++) {
            // Get the character
            ch = cyphertext.charAt(i);
            // Normalize it to uppercase
            ch = normalize(ch);
            // Calc the plaintext
            if (isKey(ch))
            {
                p[i] = alphabet[ch - 'A'];
                if (Character.isLowerCase(cyphertext.charAt(i)))
                    p[i] = Character.toLowerCase(p[i]);
                if (p[i] == ' ')
                    p[i] = '_';
            } else {
                p[i] = cyphertext.charAt(i);
            }
        }
        plaintext = new String(p);
    }
    
    public void letterDistribution(char letters[], int counts[])
    {
        char letter;
        int max = cyphertext.length();
        for (char c = 'A'; c <= 'Z'; c++)
        {
            letters[c-'A'] = c;
            counts[c-'A'] = 0;
        }
        for (int i=0; i<max; i++) {
            letter = normalize(cyphertext.charAt(i));
            if (isKey(letter))
                counts[letter-'A']++;
        }
        // BUBBLESORT!  YEAY!
        int tempInt;
        char tempChar;
        for (int i=0; i < 26; i++)
            for (int j = i; j < 26; j++)
                if (counts[j] > counts[i])
                {
                    tempChar = letters[i];
                    letters[i] = letters[j];
                    letters[j] = tempChar;
                    tempInt = counts[i];
                    counts[i] = counts[j];
                    counts[j] = tempInt;
                }
    }
    
    private Pair[] heapsort(Pair[] array) {
        int n = array.length;
        int i = n/2;
        int parent, child;
        Pair t;
        for (;;) {
            if (i > 0) {
                i--;
                t = array[i];
            } else {
                n--;
                if (n == 0)
                    return array;
                t = array[n];
                array[n] = array[0];
            }
            parent = i;
            child = i * 2 + 1;
            while (child < n) {
                if (child + 1 < n && array[child + 1].count > array[child].count) {
                    child++;
                }
                if (array[child].count > t.count) {
                    array[parent] = array[child];
                    parent = child;
                    child = parent * 2 + 1;
                } else {
                    break;
                }
            }
            array[parent] = t;
        }
    }
    
    public Pair[] wordDistribution() {
        Pair result[];
        Hashtable wordCounts = new Hashtable();
        String theWord = "";
        int max = cyphertext.length();
        int counter;
        // Calculate words in a hash table
        for (int i=0; i<max; i++) {
            char c = Character.toLowerCase(cyphertext.charAt(i));
            if (    ((c >= 'A') && (c <= 'Z'))
                 || ((c >= 'a') && (c <= 'z'))
                 || (c == '\'')
               ) {
                theWord += c;
            } else if (theWord.length() > 0) {
                Integer tempInt = (Integer) wordCounts.get(theWord);
                if (tempInt == null) {
                    tempInt = new Integer(1);
                } else {
                    tempInt = new Integer(tempInt.intValue() + 1);
                    wordCounts.remove(theWord);
                }
                wordCounts.put(theWord, tempInt);
                theWord = "";
            }
        }
        // Convert to ordered array
        counter = 0;
        result = new Pair[wordCounts.keySet().size()];
        for (Iterator i = wordCounts.keySet().iterator(); i.hasNext(); counter++) {
            String key = (String) i.next();
            int value = ((Integer) wordCounts.get(key)).intValue();
            result[counter] = new Pair(key, value);
        }
        // Heapsort array
        if (result.length > 1) {
            result = heapsort(result);
            // Reverse it
            for (int i=0; i < (result.length / 2); i++) {
                Pair temp = result[i];
                result[i] = result[result.length - i - 1];
                result[result.length - i - 1] = temp;
            }
        }
        return result;
    }

    public String getRot() {
        StringBuffer result = new StringBuffer();
        String tempMessage = new String(cyphertext);
        int max;
        tempMessage.replaceAll("\r", "");
        tempMessage.replaceAll("\n", "");
        max = tempMessage.length();
        for (int i=1; i<=25; i++) {
            String rotValue = "" + i;
            if (rotValue.length() == 1)
                rotValue = "0" + rotValue;
            result.append(rotValue);
            result.append(": ");
            for (int j=0; j<max; j++) {
                int c = tempMessage.charAt(j);
                int cu = Character.toUpperCase((char) c);
                if ((cu >= 'A') && (cu <= 'Z')) {
                    cu = cu - 'A';
                    cu = cu + i;
                    cu = cu % 26;
                    cu = cu + 'A';
                    if ((c >= 'A') && (c <= 'Z'))
                        cu = cu + 'A';
                    else 
                        cu = cu + 'a';
                    result.append((char) cu);
                } else {
                    result.append((char) c);
                }
            }
            result.append("\n");
        }
        return result.toString();
    }

    private char normalize(char letter) {
        if ( (letter >= 'a') && (letter <= 'z') )
            return (char) (letter - ('a' - 'A'));
        else
            return letter;
    }

    private boolean isKey(char letter) {
        return (letter >= 'A') && (letter <= 'Z');
    }
    
    public String getFormattedAlphabet() {
        char used[] = new char[26];
        boolean duplicates = false;
        StringBuffer result = new StringBuffer();
        result.append("Cyphertext: ");
        for (char c='A'; c<='Z'; c++)
            result.append(c);
        result.append("\nPlaintext:  ");
        for (int i=0; i<26; i++)
        {
            char c = alphabet[i];
            result.append(c);
            if (isKey(c)) {
                if (used[c - 'A'] == 1)
                    duplicates = true;
                else
                    used[c - 'A'] = 1;
            }
        }
        if (duplicates)
            result.append(" <-- Warning: duplicates");
        return result.toString();
    }

    public String getFormattedMessage(boolean asHtml) {
        StringBuffer result = new StringBuffer();
        int here=0, newHere=0, max;
        String line;
        
        calcPlaintext();
        max = cyphertext.length();
        if (asHtml)
            result.append("<html><head></head><body text=#000000 bgcolor=#ffffff><code>\n");
        // Traverse the string, one line at a time
        while (here != max)
        {
            int pos;
            // Print plaintext line
            line = plaintext.substring(here);
            pos = line.indexOf('\n');
            if (pos > 0)
                line = line.substring(0, pos);
            if (asHtml)
                result.append("<font color=\"#000000\">");
            result.append(line);
            if (asHtml)
                result.append("</font><br>");
            result.append("\n");
            // Print cyphertext line
            line = cyphertext.substring(here);
            pos = line.indexOf('\n');
            if (pos > 0)
                line = line.substring(0, pos);
            if (asHtml)
                result.append("<font color=\"#999999\">");
            result.append(line);
            if (asHtml)
                result.append("</font><br>");
            result.append("\n");
            // Move on to next line
            newHere = cyphertext.indexOf('\n', here);
            // If no more newlines, ensure we are at the end
            if (newHere == -1)
                here = max;
            else
                here = newHere + 1;
        }
        if (asHtml)
            result.append("</code></body></html>\n");
        return result.toString();
    }
    
    public class Pair {
        public String word;
        public int count;
        public Pair(String s, int i) {this.word = s; this.count = i;}
    }
}
